package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.Arrays;
import java.util.Comparator;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.core.compiler.InvalidInputException;
import org.eclipse.jdt.internal.compiler.ast.LambdaExpression;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.NullAnnotationMatching;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationHolder;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.CaptureBinding;
import org.eclipse.jdt.internal.compiler.lookup.CompilationUnitScope;
import org.eclipse.jdt.internal.compiler.lookup.FieldBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.PackageBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.util.SimpleLookupTable;

public abstract class ReferenceBinding extends TypeBinding {
  public char[][] compoundName;
  
  public char[] sourceName;
  
  public int modifiers;
  
  public PackageBinding fPackage;
  
  char[] fileName;
  
  char[] constantPoolName;
  
  char[] signature;
  
  private SimpleLookupTable compatibleCache;
  
  int typeBits;
  
  protected MethodBinding[] singleAbstractMethod;
  
  public static final org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding LUB_GENERIC = (org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)new Object();
  
  private static final Comparator<FieldBinding> FIELD_COMPARATOR = (Comparator<FieldBinding>)new Object();
  
  private static final Comparator<MethodBinding> METHOD_COMPARATOR = (Comparator<MethodBinding>)new Object();
  
  protected static ProblemMethodBinding samProblemBinding = new ProblemMethodBinding(TypeConstants.ANONYMOUS_METHOD, null, 17);
  
  public ReferenceBinding(org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding prototype) {
    super(prototype);
    this.compoundName = prototype.compoundName;
    this.sourceName = prototype.sourceName;
    this.modifiers = prototype.modifiers;
    this.fPackage = prototype.fPackage;
    this.fileName = prototype.fileName;
    this.constantPoolName = prototype.constantPoolName;
    this.signature = prototype.signature;
    this.compatibleCache = prototype.compatibleCache;
    this.typeBits = prototype.typeBits;
    this.singleAbstractMethod = prototype.singleAbstractMethod;
  }
  
  public ReferenceBinding() {}
  
  public static FieldBinding binarySearch(char[] name, FieldBinding[] sortedFields) {
    if (sortedFields == null)
      return null; 
    int max = sortedFields.length;
    if (max == 0)
      return null; 
    int left = 0, right = max - 1, nameLength = name.length;
    int mid = 0;
    while (left <= right) {
      mid = left + (right - left) / 2;
      char[] midName;
      int compare = compare(name, midName = (sortedFields[mid]).name, nameLength, midName.length);
      if (compare < 0) {
        right = mid - 1;
        continue;
      } 
      if (compare > 0) {
        left = mid + 1;
        continue;
      } 
      return sortedFields[mid];
    } 
    return null;
  }
  
  public static long binarySearch(char[] selector, MethodBinding[] sortedMethods) {
    if (sortedMethods == null)
      return -1L; 
    int max = sortedMethods.length;
    if (max == 0)
      return -1L; 
    int left = 0, right = max - 1, selectorLength = selector.length;
    int mid = 0;
    while (left <= right) {
      mid = left + (right - left) / 2;
      char[] midSelector;
      int compare = compare(selector, midSelector = (sortedMethods[mid]).selector, selectorLength, midSelector.length);
      if (compare < 0) {
        right = mid - 1;
        continue;
      } 
      if (compare > 0) {
        left = mid + 1;
        continue;
      } 
      int start = mid, end = mid;
      for (; start > left && CharOperation.equals((sortedMethods[start - 1]).selector, selector); start--);
      for (; end < right && CharOperation.equals((sortedMethods[end + 1]).selector, selector); end++);
      return start + (end << 32L);
    } 
    return -1L;
  }
  
  static int compare(char[] str1, char[] str2, int len1, int len2) {
    int n = Math.min(len1, len2);
    int i = 0;
    while (n-- != 0) {
      char c1 = str1[i];
      char c2 = str2[i++];
      if (c1 != c2)
        return c1 - c2; 
    } 
    return len1 - len2;
  }
  
  public static void sortFields(FieldBinding[] sortedFields, int left, int right) {
    Arrays.sort(sortedFields, left, right, FIELD_COMPARATOR);
  }
  
  public static void sortMethods(MethodBinding[] sortedMethods, int left, int right) {
    Arrays.sort(sortedMethods, left, right, METHOD_COMPARATOR);
  }
  
  public FieldBinding[] availableFields() {
    return fields();
  }
  
  public MethodBinding[] availableMethods() {
    return methods();
  }
  
  public boolean canBeInstantiated() {
    return ((this.modifiers & 0x6600) == 0);
  }
  
  public boolean canBeSeenBy(PackageBinding invocationPackage) {
    if (isPublic())
      return true; 
    if (isPrivate())
      return false; 
    return (invocationPackage == this.fPackage);
  }
  
  public boolean canBeSeenBy(org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding receiverType, org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding invocationType) {
    // Byte code:
    //   0: aload_0
    //   1: invokevirtual isPublic : ()Z
    //   4: ifeq -> 9
    //   7: iconst_1
    //   8: ireturn
    //   9: aload_2
    //   10: aload_0
    //   11: invokestatic equalsEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   14: ifeq -> 27
    //   17: aload_2
    //   18: aload_1
    //   19: invokestatic equalsEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   22: ifeq -> 27
    //   25: iconst_1
    //   26: ireturn
    //   27: aload_0
    //   28: invokevirtual isProtected : ()Z
    //   31: ifeq -> 111
    //   34: aload_2
    //   35: aload_0
    //   36: invokestatic equalsEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   39: ifeq -> 44
    //   42: iconst_1
    //   43: ireturn
    //   44: aload_2
    //   45: getfield fPackage : Lorg/eclipse/jdt/internal/compiler/lookup/PackageBinding;
    //   48: aload_0
    //   49: getfield fPackage : Lorg/eclipse/jdt/internal/compiler/lookup/PackageBinding;
    //   52: if_acmpne -> 57
    //   55: iconst_1
    //   56: ireturn
    //   57: aload_2
    //   58: invokevirtual erasure : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   61: astore_3
    //   62: aload_0
    //   63: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   66: invokevirtual erasure : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   69: astore #4
    //   71: aload #4
    //   73: aload_2
    //   74: invokestatic equalsEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   77: ifeq -> 82
    //   80: iconst_1
    //   81: ireturn
    //   82: aload #4
    //   84: ifnonnull -> 89
    //   87: iconst_0
    //   88: ireturn
    //   89: aload_3
    //   90: aload #4
    //   92: invokevirtual findSuperTypeOriginatingFrom : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   95: ifnull -> 100
    //   98: iconst_1
    //   99: ireturn
    //   100: aload_3
    //   101: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   104: astore_3
    //   105: aload_3
    //   106: ifnonnull -> 89
    //   109: iconst_0
    //   110: ireturn
    //   111: aload_0
    //   112: invokevirtual isPrivate : ()Z
    //   115: ifeq -> 278
    //   118: aload_1
    //   119: aload_0
    //   120: invokestatic equalsEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   123: ifne -> 196
    //   126: aload_1
    //   127: aload_0
    //   128: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   131: invokestatic equalsEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   134: ifne -> 196
    //   137: aload_1
    //   138: invokevirtual isTypeVariable : ()Z
    //   141: ifeq -> 194
    //   144: aload_1
    //   145: checkcast org/eclipse/jdt/internal/compiler/lookup/TypeVariableBinding
    //   148: astore_3
    //   149: aload_3
    //   150: getfield environment : Lorg/eclipse/jdt/internal/compiler/lookup/LookupEnvironment;
    //   153: getfield globalOptions : Lorg/eclipse/jdt/internal/compiler/impl/CompilerOptions;
    //   156: getfield complianceLevel : J
    //   159: ldc2_w 3276800
    //   162: lcmp
    //   163: ifgt -> 194
    //   166: aload_3
    //   167: aload_0
    //   168: invokevirtual erasure : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   171: invokevirtual isErasureBoundTo : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   174: ifne -> 196
    //   177: aload_3
    //   178: aload_0
    //   179: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   182: invokevirtual erasure : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   185: invokevirtual isErasureBoundTo : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   188: ifeq -> 194
    //   191: goto -> 196
    //   194: iconst_0
    //   195: ireturn
    //   196: aload_2
    //   197: aload_0
    //   198: invokestatic notEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   201: ifeq -> 276
    //   204: aload_2
    //   205: astore_3
    //   206: aload_3
    //   207: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   210: astore #4
    //   212: goto -> 225
    //   215: aload #4
    //   217: astore_3
    //   218: aload #4
    //   220: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   223: astore #4
    //   225: aload #4
    //   227: ifnonnull -> 215
    //   230: aload_0
    //   231: invokevirtual erasure : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   234: checkcast org/eclipse/jdt/internal/compiler/lookup/ReferenceBinding
    //   237: astore #5
    //   239: aload #5
    //   241: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   244: astore #4
    //   246: goto -> 260
    //   249: aload #4
    //   251: astore #5
    //   253: aload #4
    //   255: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   258: astore #4
    //   260: aload #4
    //   262: ifnonnull -> 249
    //   265: aload_3
    //   266: aload #5
    //   268: invokestatic notEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   271: ifeq -> 276
    //   274: iconst_0
    //   275: ireturn
    //   276: iconst_1
    //   277: ireturn
    //   278: aload_2
    //   279: getfield fPackage : Lorg/eclipse/jdt/internal/compiler/lookup/PackageBinding;
    //   282: aload_0
    //   283: getfield fPackage : Lorg/eclipse/jdt/internal/compiler/lookup/PackageBinding;
    //   286: if_acmpeq -> 291
    //   289: iconst_0
    //   290: ireturn
    //   291: aload_1
    //   292: astore_3
    //   293: aload_0
    //   294: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   297: ifnonnull -> 304
    //   300: aload_0
    //   301: goto -> 308
    //   304: aload_0
    //   305: invokevirtual enclosingType : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   308: invokevirtual original : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   311: astore #4
    //   313: aload_3
    //   314: invokevirtual isCapture : ()Z
    //   317: ifeq -> 337
    //   320: aload #4
    //   322: aload_3
    //   323: invokevirtual erasure : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   326: invokevirtual original : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   329: invokestatic equalsEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   332: ifeq -> 351
    //   335: iconst_1
    //   336: ireturn
    //   337: aload #4
    //   339: aload_3
    //   340: invokevirtual original : ()Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   343: invokestatic equalsEquals : (Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;)Z
    //   346: ifeq -> 351
    //   349: iconst_1
    //   350: ireturn
    //   351: aload_3
    //   352: getfield fPackage : Lorg/eclipse/jdt/internal/compiler/lookup/PackageBinding;
    //   355: astore #5
    //   357: aload #5
    //   359: ifnull -> 373
    //   362: aload #5
    //   364: aload_0
    //   365: getfield fPackage : Lorg/eclipse/jdt/internal/compiler/lookup/PackageBinding;
    //   368: if_acmpeq -> 373
    //   371: iconst_0
    //   372: ireturn
    //   373: aload_3
    //   374: invokevirtual superclass : ()Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   377: dup
    //   378: astore_3
    //   379: ifnonnull -> 313
    //   382: iconst_0
    //   383: ireturn
    // Line number table:
    //   Java source line number -> byte code offset
    //   #263	-> 0
    //   #265	-> 9
    //   #267	-> 27
    //   #273	-> 34
    //   #274	-> 44
    //   #276	-> 57
    //   #277	-> 62
    //   #278	-> 71
    //   #279	-> 82
    //   #282	-> 89
    //   #284	-> 100
    //   #285	-> 105
    //   #286	-> 109
    //   #289	-> 111
    //   #293	-> 118
    //   #295	-> 137
    //   #296	-> 144
    //   #297	-> 149
    //   #298	-> 191
    //   #300	-> 194
    //   #304	-> 196
    //   #305	-> 204
    //   #306	-> 206
    //   #307	-> 212
    //   #308	-> 215
    //   #309	-> 218
    //   #307	-> 225
    //   #312	-> 230
    //   #313	-> 239
    //   #314	-> 246
    //   #315	-> 249
    //   #316	-> 253
    //   #314	-> 260
    //   #318	-> 265
    //   #320	-> 276
    //   #324	-> 278
    //   #326	-> 291
    //   #327	-> 293
    //   #329	-> 313
    //   #330	-> 320
    //   #332	-> 337
    //   #334	-> 351
    //   #336	-> 357
    //   #337	-> 373
    //   #338	-> 382
    // Local variable table:
    //   start	length	slot	name	descriptor
    //   0	384	0	this	Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   0	384	1	receiverType	Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   0	384	2	invocationType	Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   62	49	3	currentType	Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   71	40	4	declaringClass	Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   149	45	3	typeVariable	Lorg/eclipse/jdt/internal/compiler/lookup/TypeVariableBinding;
    //   206	70	3	outerInvocationType	Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   212	64	4	temp	Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   239	37	5	outerDeclaringClass	Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   293	91	3	currentType	Lorg/eclipse/jdt/internal/compiler/lookup/ReferenceBinding;
    //   313	71	4	originalDeclaringClass	Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   357	16	5	currentPackage	Lorg/eclipse/jdt/internal/compiler/lookup/PackageBinding;
  }
  
  public boolean canBeSeenBy(Scope scope) {
    if (isPublic())
      return true; 
    SourceTypeBinding invocationType = scope.enclosingSourceType();
    if (TypeBinding.equalsEquals((TypeBinding)invocationType, this))
      return true; 
    if (invocationType == null)
      return (!isPrivate() && scope.getCurrentPackage() == this.fPackage); 
    if (isProtected()) {
      if (invocationType.fPackage == this.fPackage)
        return true; 
      TypeBinding declaringClass = enclosingType();
      if (declaringClass == null)
        return false; 
      declaringClass = declaringClass.erasure();
      TypeBinding currentType = invocationType.erasure();
      while (true) {
        if (TypeBinding.equalsEquals(declaringClass, (TypeBinding)invocationType))
          return true; 
        if (currentType.findSuperTypeOriginatingFrom(declaringClass) != null)
          return true; 
        currentType = currentType.enclosingType();
        if (currentType == null)
          return false; 
      } 
    } 
    if (isPrivate()) {
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding referenceBinding1;
      SourceTypeBinding sourceTypeBinding = invocationType;
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding temp = sourceTypeBinding.enclosingType();
      while (temp != null) {
        referenceBinding1 = temp;
        temp = temp.enclosingType();
      } 
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding outerDeclaringClass = (org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)erasure();
      temp = outerDeclaringClass.enclosingType();
      while (temp != null) {
        outerDeclaringClass = temp;
        temp = temp.enclosingType();
      } 
      return TypeBinding.equalsEquals(referenceBinding1, outerDeclaringClass);
    } 
    return (invocationType.fPackage == this.fPackage);
  }
  
  public char[] computeGenericTypeSignature(TypeVariableBinding[] typeVariables) {
    boolean isMemberOfGeneric = (isMemberType() && ((enclosingType()).modifiers & 0x40000000) != 0);
    if (typeVariables == Binding.NO_TYPE_VARIABLES && !isMemberOfGeneric)
      return signature(); 
    StringBuffer sig = new StringBuffer(10);
    if (isMemberOfGeneric) {
      char[] typeSig = enclosingType().genericTypeSignature();
      sig.append(typeSig, 0, typeSig.length - 1);
      sig.append('.');
      sig.append(this.sourceName);
    } else {
      char[] typeSig = signature();
      sig.append(typeSig, 0, typeSig.length - 1);
    } 
    if (typeVariables == Binding.NO_TYPE_VARIABLES) {
      sig.append(';');
    } else {
      sig.append('<');
      for (int i = 0, length = typeVariables.length; i < length; i++)
        sig.append(typeVariables[i].genericTypeSignature()); 
      sig.append(">;");
    } 
    int sigLength = sig.length();
    char[] result = new char[sigLength];
    sig.getChars(0, sigLength, result, 0);
    return result;
  }
  
  public void computeId() {
    char[] packageName;
    char[] typeName;
    switch (this.compoundName.length) {
      case 3:
        packageName = this.compoundName[0];
        switch (packageName.length) {
          case 3:
            if (CharOperation.equals(TypeConstants.ORG_JUNIT_ASSERT, this.compoundName))
              this.id = 70; 
            return;
          case 4:
            if (!CharOperation.equals(TypeConstants.JAVA, packageName))
              return; 
            break;
          case 5:
            switch (packageName[1]) {
              case 'a':
                if (CharOperation.equals(TypeConstants.JAVAX_ANNOTATION_INJECT_INJECT, this.compoundName))
                  this.id = 80; 
                return;
              case 'u':
                if (CharOperation.equals(TypeConstants.JUNIT_FRAMEWORK_ASSERT, this.compoundName))
                  this.id = 69; 
                return;
            } 
            return;
          default:
            return;
        } 
        packageName = this.compoundName[1];
        if (packageName.length == 0)
          return; 
        typeName = this.compoundName[2];
        if (typeName.length == 0)
          return; 
        if (!CharOperation.equals(TypeConstants.LANG, this.compoundName[1])) {
          switch (packageName[0]) {
            case 'i':
              if (CharOperation.equals(packageName, TypeConstants.IO))
                switch (typeName[0]) {
                  case 'C':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_IO_CLOSEABLE[2]))
                      this.typeBits |= 0x2; 
                    return;
                  case 'E':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_IO_EXTERNALIZABLE[2]))
                      this.id = 56; 
                    return;
                  case 'I':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_IO_IOEXCEPTION[2]))
                      this.id = 58; 
                    return;
                  case 'O':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_IO_OBJECTSTREAMEXCEPTION[2]))
                      this.id = 57; 
                    return;
                  case 'P':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_IO_PRINTSTREAM[2]))
                      this.id = 53; 
                    return;
                  case 'S':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_IO_SERIALIZABLE[2]))
                      this.id = 37; 
                    return;
                }  
              return;
            case 'u':
              if (CharOperation.equals(packageName, TypeConstants.UTIL))
                switch (typeName[0]) {
                  case 'C':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_UTIL_COLLECTION[2]))
                      this.id = 59; 
                    return;
                  case 'I':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_UTIL_ITERATOR[2]))
                      this.id = 39; 
                    return;
                  case 'O':
                    if (CharOperation.equals(typeName, TypeConstants.JAVA_UTIL_OBJECTS[2]))
                      this.id = 74; 
                    return;
                }  
              return;
          } 
          return;
        } 
        switch (typeName[0]) {
          case 'A':
            switch (typeName.length) {
              case 13:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_AUTOCLOSEABLE[2])) {
                  this.id = 62;
                  this.typeBits |= 0x1;
                } 
                return;
              case 14:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ASSERTIONERROR[2]))
                  this.id = 35; 
                return;
            } 
            return;
          case 'B':
            switch (typeName.length) {
              case 4:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_BYTE[2]))
                  this.id = 26; 
                return;
              case 7:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_BOOLEAN[2]))
                  this.id = 33; 
                return;
            } 
            return;
          case 'C':
            switch (typeName.length) {
              case 5:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_CLASS[2]))
                  this.id = 16; 
                return;
              case 9:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_CHARACTER[2])) {
                  this.id = 28;
                } else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_CLONEABLE[2])) {
                  this.id = 36;
                } 
                return;
              case 22:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_CLASSNOTFOUNDEXCEPTION[2]))
                  this.id = 23; 
                return;
            } 
            return;
          case 'D':
            switch (typeName.length) {
              case 6:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_DOUBLE[2]))
                  this.id = 32; 
                return;
              case 10:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_DEPRECATED[2]))
                  this.id = 44; 
                return;
            } 
            return;
          case 'E':
            switch (typeName.length) {
              case 4:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ENUM[2]))
                  this.id = 41; 
                return;
              case 5:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ERROR[2]))
                  this.id = 19; 
                return;
              case 9:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_EXCEPTION[2]))
                  this.id = 25; 
                return;
            } 
            return;
          case 'F':
            if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_FLOAT[2])) {
              this.id = 31;
            } else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_FUNCTIONAL_INTERFACE[2])) {
              this.id = 77;
            } 
            return;
          case 'I':
            switch (typeName.length) {
              case 7:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_INTEGER[2]))
                  this.id = 29; 
                return;
              case 8:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ITERABLE[2]))
                  this.id = 38; 
                return;
              case 24:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ILLEGALARGUMENTEXCEPTION[2]))
                  this.id = 42; 
                return;
            } 
            return;
          case 'L':
            if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_LONG[2]))
              this.id = 30; 
            return;
          case 'N':
            if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_NOCLASSDEFERROR[2]))
              this.id = 22; 
            return;
          case 'O':
            switch (typeName.length) {
              case 6:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_OBJECT[2]))
                  this.id = 1; 
                return;
              case 8:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_OVERRIDE[2]))
                  this.id = 47; 
                return;
            } 
            return;
          case 'R':
            if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_RUNTIMEEXCEPTION[2]))
              this.id = 24; 
            break;
          case 'S':
            switch (typeName.length) {
              case 5:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_SHORT[2]))
                  this.id = 27; 
                return;
              case 6:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_STRING[2])) {
                  this.id = 11;
                } else if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_SYSTEM[2])) {
                  this.id = 18;
                } 
                return;
              case 11:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_SAFEVARARGS[2]))
                  this.id = 60; 
                return;
              case 12:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_STRINGBUFFER[2]))
                  this.id = 17; 
                return;
              case 13:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_STRINGBUILDER[2]))
                  this.id = 40; 
                return;
              case 16:
                if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_SUPPRESSWARNINGS[2]))
                  this.id = 49; 
                return;
            } 
            return;
          case 'T':
            if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_THROWABLE[2]))
              this.id = 21; 
            return;
          case 'V':
            if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_VOID[2]))
              this.id = 34; 
            return;
        } 
        break;
      case 4:
        if (CharOperation.equals(TypeConstants.COM_GOOGLE_INJECT_INJECT, this.compoundName)) {
          this.id = 81;
          return;
        } 
        if (!CharOperation.equals(TypeConstants.JAVA, this.compoundName[0]))
          return; 
        packageName = this.compoundName[1];
        if (packageName.length == 0)
          return; 
        packageName = this.compoundName[2];
        if (packageName.length == 0)
          return; 
        typeName = this.compoundName[3];
        if (typeName.length == 0)
          return; 
        switch (packageName[0]) {
          case 'a':
            if (CharOperation.equals(packageName, TypeConstants.ANNOTATION))
              switch (typeName[0]) {
                case 'A':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_ANNOTATION[3]))
                    this.id = 43; 
                  return;
                case 'D':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_DOCUMENTED[3]))
                    this.id = 45; 
                  return;
                case 'E':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_ELEMENTTYPE[3]))
                    this.id = 52; 
                  return;
                case 'I':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_INHERITED[3]))
                    this.id = 46; 
                  return;
                case 'R':
                  switch (typeName.length) {
                    case 9:
                      if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_RETENTION[3]))
                        this.id = 48; 
                      return;
                    case 10:
                      if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_REPEATABLE[3]))
                        this.id = 90; 
                      return;
                    case 15:
                      if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_RETENTIONPOLICY[3]))
                        this.id = 51; 
                      return;
                  } 
                  return;
                case 'T':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_ANNOTATION_TARGET[3]))
                    this.id = 50; 
                  return;
              }  
            return;
          case 'i':
            if (CharOperation.equals(packageName, TypeConstants.INVOKE)) {
              if (typeName.length == 0)
                return; 
              switch (typeName[0]) {
                case 'M':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_INVOKE_METHODHANDLE_$_POLYMORPHICSIGNATURE[3]))
                    this.id = 61; 
                  return;
              } 
            } 
            return;
          case 'r':
            if (CharOperation.equals(packageName, TypeConstants.REFLECT))
              switch (typeName[0]) {
                case 'C':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_REFLECT_CONSTRUCTOR[2]))
                    this.id = 20; 
                  return;
                case 'F':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_REFLECT_FIELD[2]))
                    this.id = 54; 
                  return;
                case 'M':
                  if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_REFLECT_METHOD[2]))
                    this.id = 55; 
                  return;
              }  
            return;
        } 
        break;
      case 5:
        packageName = this.compoundName[0];
        switch (packageName[0]) {
          case 'j':
            if (!CharOperation.equals(TypeConstants.JAVA, this.compoundName[0]))
              return; 
            packageName = this.compoundName[1];
            if (packageName.length == 0)
              return; 
            if (CharOperation.equals(TypeConstants.LANG, packageName)) {
              packageName = this.compoundName[2];
              if (packageName.length == 0)
                return; 
              switch (packageName[0]) {
                case 'i':
                  if (CharOperation.equals(packageName, TypeConstants.INVOKE)) {
                    char[] memberTypeName;
                    typeName = this.compoundName[3];
                    if (typeName.length == 0)
                      return; 
                    switch (typeName[0]) {
                      case 'M':
                        memberTypeName = this.compoundName[4];
                        if (memberTypeName.length == 0)
                          return; 
                        if (CharOperation.equals(typeName, TypeConstants.JAVA_LANG_INVOKE_METHODHANDLE_POLYMORPHICSIGNATURE[3]) && 
                          CharOperation.equals(memberTypeName, TypeConstants.JAVA_LANG_INVOKE_METHODHANDLE_POLYMORPHICSIGNATURE[4]))
                          this.id = 61; 
                        return;
                    } 
                  } 
                  return;
              } 
              return;
            } 
            return;
          case 'o':
            if (!CharOperation.equals(TypeConstants.ORG, this.compoundName[0]))
              return; 
            packageName = this.compoundName[1];
            if (packageName.length == 0)
              return; 
            switch (packageName[0]) {
              case 'e':
                if (CharOperation.equals(TypeConstants.ECLIPSE, packageName)) {
                  packageName = this.compoundName[2];
                  if (packageName.length == 0)
                    return; 
                  switch (packageName[0]) {
                    case 'c':
                      if (CharOperation.equals(packageName, TypeConstants.CORE)) {
                        char[] memberTypeName;
                        typeName = this.compoundName[3];
                        if (typeName.length == 0)
                          return; 
                        switch (typeName[0]) {
                          case 'r':
                            memberTypeName = this.compoundName[4];
                            if (memberTypeName.length == 0)
                              return; 
                            if (CharOperation.equals(typeName, TypeConstants.ORG_ECLIPSE_CORE_RUNTIME_ASSERT[3]) && 
                              CharOperation.equals(memberTypeName, TypeConstants.ORG_ECLIPSE_CORE_RUNTIME_ASSERT[4]))
                              this.id = 68; 
                            return;
                        } 
                      } 
                      return;
                  } 
                  return;
                } 
                return;
              case 'a':
                if (CharOperation.equals(TypeConstants.APACHE, packageName) && 
                  CharOperation.equals(TypeConstants.COMMONS, this.compoundName[2]))
                  if (CharOperation.equals(TypeConstants.ORG_APACHE_COMMONS_LANG_VALIDATE, this.compoundName)) {
                    this.id = 71;
                  } else if (CharOperation.equals(TypeConstants.ORG_APACHE_COMMONS_LANG3_VALIDATE, this.compoundName)) {
                    this.id = 72;
                  }  
                return;
            } 
            return;
          case 'c':
            if (!CharOperation.equals(TypeConstants.COM, this.compoundName[0]))
              return; 
            if (CharOperation.equals(TypeConstants.COM_GOOGLE_COMMON_BASE_PRECONDITIONS, this.compoundName))
              this.id = 73; 
            return;
        } 
        break;
      case 6:
        if (CharOperation.equals(TypeConstants.ORG, this.compoundName[0])) {
          if (CharOperation.equals(TypeConstants.SPRING, this.compoundName[1])) {
            if (CharOperation.equals(TypeConstants.AUTOWIRED, this.compoundName[5]) && 
              CharOperation.equals(TypeConstants.ORG_SPRING_AUTOWIRED, this.compoundName))
              this.id = 82; 
            return;
          } 
          if (!CharOperation.equals(TypeConstants.JDT, this.compoundName[2]) || !CharOperation.equals(TypeConstants.ITYPEBINDING, this.compoundName[5]))
            return; 
          if (CharOperation.equals(TypeConstants.ORG_ECLIPSE_JDT_CORE_DOM_ITYPEBINDING, this.compoundName))
            this.typeBits |= 0x10; 
        } 
        break;
      case 7:
        if (!CharOperation.equals(TypeConstants.JDT, this.compoundName[2]) || !CharOperation.equals(TypeConstants.TYPEBINDING, this.compoundName[6]))
          return; 
        if (CharOperation.equals(TypeConstants.ORG_ECLIPSE_JDT_INTERNAL_COMPILER_LOOKUP_TYPEBINDING, this.compoundName))
          this.typeBits |= 0x10; 
        break;
    } 
  }
  
  public void computeId(LookupEnvironment environment) {
    environment.getUnannotatedType(this);
  }
  
  public char[] computeUniqueKey(boolean isLeaf) {
    if (!isLeaf)
      return signature(); 
    return genericTypeSignature();
  }
  
  public char[] constantPoolName() {
    if (this.constantPoolName != null)
      return this.constantPoolName; 
    return this.constantPoolName = CharOperation.concatWith(this.compoundName, '/');
  }
  
  public String debugName() {
    return (this.compoundName != null) ? (hasTypeAnnotations() ? annotatedDebugName() : new String(readableName())) : "UNNAMED TYPE";
  }
  
  public int depth() {
    int depth = 0;
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding current = this;
    while ((current = current.enclosingType()) != null)
      depth++; 
    return depth;
  }
  
  public boolean detectAnnotationCycle() {
    if ((this.tagBits & 0x100000000L) != 0L)
      return false; 
    if ((this.tagBits & 0x80000000L) != 0L)
      return true; 
    this.tagBits |= 0x80000000L;
    MethodBinding[] currentMethods = methods();
    boolean inCycle = false;
    for (int i = 0, l = currentMethods.length; i < l; i++) {
      TypeBinding returnType = (currentMethods[i]).returnType.leafComponentType().erasure();
      if (TypeBinding.equalsEquals(this, returnType)) {
        if (this instanceof SourceTypeBinding) {
          MethodDeclaration decl = (MethodDeclaration)currentMethods[i].sourceMethod();
          ((SourceTypeBinding)this).scope.problemReporter().annotationCircularity(this, this, (decl != null) ? decl.returnType : null);
        } 
      } else if (returnType.isAnnotationType() && ((org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)returnType).detectAnnotationCycle()) {
        if (this instanceof SourceTypeBinding) {
          MethodDeclaration decl = (MethodDeclaration)currentMethods[i].sourceMethod();
          ((SourceTypeBinding)this).scope.problemReporter().annotationCircularity(this, returnType, (decl != null) ? decl.returnType : null);
        } 
        inCycle = true;
      } 
    } 
    if (inCycle)
      return true; 
    this.tagBits |= 0x100000000L;
    return false;
  }
  
  public final org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding enclosingTypeAt(int relativeDepth) {
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding current = this;
    while (relativeDepth-- > 0 && current != null)
      current = current.enclosingType(); 
    return current;
  }
  
  public int enumConstantCount() {
    int count = 0;
    FieldBinding[] fields = fields();
    for (int i = 0, length = fields.length; i < length; i++) {
      if (((fields[i]).modifiers & 0x4000) != 0)
        count++; 
    } 
    return count;
  }
  
  public int fieldCount() {
    return (fields()).length;
  }
  
  public FieldBinding[] fields() {
    return Binding.NO_FIELDS;
  }
  
  public final int getAccessFlags() {
    return this.modifiers & 0xFFFF;
  }
  
  public AnnotationBinding[] getAnnotations() {
    return retrieveAnnotations((Binding)this);
  }
  
  public long getAnnotationTagBits() {
    return this.tagBits;
  }
  
  public int getEnclosingInstancesSlotSize() {
    if (isStatic())
      return 0; 
    return (enclosingType() == null) ? 0 : 1;
  }
  
  public MethodBinding getExactConstructor(TypeBinding[] argumentTypes) {
    return null;
  }
  
  public MethodBinding getExactMethod(char[] selector, TypeBinding[] argumentTypes, CompilationUnitScope refScope) {
    return null;
  }
  
  public FieldBinding getField(char[] fieldName, boolean needResolve) {
    return null;
  }
  
  public char[] getFileName() {
    return this.fileName;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding getMemberType(char[] typeName) {
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] memberTypes = memberTypes();
    for (int i = memberTypes.length; --i >= 0;) {
      if (CharOperation.equals((memberTypes[i]).sourceName, typeName))
        return memberTypes[i]; 
    } 
    return null;
  }
  
  public MethodBinding[] getMethods(char[] selector) {
    return Binding.NO_METHODS;
  }
  
  public MethodBinding[] getMethods(char[] selector, int suggestedParameterLength) {
    return getMethods(selector);
  }
  
  public int getOuterLocalVariablesSlotSize() {
    return 0;
  }
  
  public PackageBinding getPackage() {
    return this.fPackage;
  }
  
  public TypeVariableBinding getTypeVariable(char[] variableName) {
    TypeVariableBinding[] typeVariables = typeVariables();
    for (int i = typeVariables.length; --i >= 0;) {
      if (CharOperation.equals((typeVariables[i]).sourceName, variableName))
        return typeVariables[i]; 
    } 
    return null;
  }
  
  public int hashCode() {
    return (this.compoundName == null || this.compoundName.length == 0) ? 
      super.hashCode() : 
      CharOperation.hashCode(this.compoundName[this.compoundName.length - 1]);
  }
  
  public boolean hasIncompatibleSuperType(org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding otherType) {
    if (TypeBinding.equalsEquals(this, otherType))
      return false; 
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] interfacesToVisit = null;
    int nextPosition = 0;
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding currentType = this;
    do {
      TypeBinding match = otherType.findSuperTypeOriginatingFrom(currentType);
      if (match != null && match.isProvablyDistinct(currentType))
        return true; 
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if (itsInterfaces == null || itsInterfaces == Binding.NO_SUPERINTERFACES)
        continue; 
      if (interfacesToVisit == null) {
        interfacesToVisit = itsInterfaces;
        nextPosition = interfacesToVisit.length;
      } else {
        int itsLength = itsInterfaces.length;
        if (nextPosition + itsLength >= interfacesToVisit.length)
          System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
        for (int a = 0; a < itsLength; a++) {
          org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding next = itsInterfaces[a];
          int b = 0;
          while (true) {
            if (b >= nextPosition) {
              interfacesToVisit[nextPosition++] = next;
              break;
            } 
            if (TypeBinding.equalsEquals(next, interfacesToVisit[b]))
              break; 
            b++;
          } 
        } 
      } 
    } while ((currentType = currentType.superclass()) != null);
    for (int i = 0; i < nextPosition; i++) {
      currentType = interfacesToVisit[i];
      if (TypeBinding.equalsEquals(currentType, otherType))
        return false; 
      TypeBinding typeBinding = otherType.findSuperTypeOriginatingFrom(currentType);
      if (typeBinding != null && typeBinding.isProvablyDistinct(currentType))
        return true; 
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if (itsInterfaces != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
        int itsLength = itsInterfaces.length;
        if (nextPosition + itsLength >= interfacesToVisit.length)
          System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
        for (int a = 0; a < itsLength; a++) {
          org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding next = itsInterfaces[a];
          int b = 0;
          while (true) {
            if (b >= nextPosition) {
              interfacesToVisit[nextPosition++] = next;
              break;
            } 
            if (TypeBinding.equalsEquals(next, interfacesToVisit[b]))
              break; 
            b++;
          } 
        } 
      } 
    } 
    return false;
  }
  
  public boolean hasMemberTypes() {
    return false;
  }
  
  boolean hasNonNullDefaultFor(int location, boolean useTypeAnnotations) {
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding currentType = this;
    while (currentType != null) {
      if (useTypeAnnotations) {
        int nullDefault = ((org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)currentType.original()).getNullDefault();
        if (nullDefault != 0)
          return ((nullDefault & location) != 0); 
      } else {
        if ((currentType.tagBits & 0x200000000000000L) != 0L)
          return true; 
        if ((currentType.tagBits & 0x400000000000000L) != 0L)
          return false; 
      } 
      currentType = currentType.enclosingType();
    } 
    if (useTypeAnnotations)
      return (((getPackage()).defaultNullness & location) != 0); 
    return ((getPackage()).defaultNullness == 1);
  }
  
  int getNullDefault() {
    return 0;
  }
  
  public boolean acceptsNonNullDefault() {
    return true;
  }
  
  public final boolean hasRestrictedAccess() {
    return ((this.modifiers & 0x40000) != 0);
  }
  
  public boolean hasNullBit(int mask) {
    return ((this.typeBits & mask) != 0);
  }
  
  public boolean implementsInterface(org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding anInterface, boolean searchHierarchy) {
    if (TypeBinding.equalsEquals(this, anInterface))
      return true; 
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] interfacesToVisit = null;
    int nextPosition = 0;
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding currentType = this;
    do {
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if (itsInterfaces == null || itsInterfaces == Binding.NO_SUPERINTERFACES)
        continue; 
      if (interfacesToVisit == null) {
        interfacesToVisit = itsInterfaces;
        nextPosition = interfacesToVisit.length;
      } else {
        int itsLength = itsInterfaces.length;
        if (nextPosition + itsLength >= interfacesToVisit.length)
          System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
        for (int a = 0; a < itsLength; a++) {
          org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding next = itsInterfaces[a];
          int b = 0;
          while (true) {
            if (b >= nextPosition) {
              interfacesToVisit[nextPosition++] = next;
              break;
            } 
            if (TypeBinding.equalsEquals(next, interfacesToVisit[b]))
              break; 
            b++;
          } 
        } 
      } 
    } while (searchHierarchy && (currentType = currentType.superclass()) != null);
    for (int i = 0; i < nextPosition; i++) {
      currentType = interfacesToVisit[i];
      if (currentType.isEquivalentTo(anInterface))
        return true; 
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] itsInterfaces = currentType.superInterfaces();
      if (itsInterfaces != null && itsInterfaces != Binding.NO_SUPERINTERFACES) {
        int itsLength = itsInterfaces.length;
        if (nextPosition + itsLength >= interfacesToVisit.length)
          System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
        for (int a = 0; a < itsLength; a++) {
          org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding next = itsInterfaces[a];
          int b = 0;
          while (true) {
            if (b >= nextPosition) {
              interfacesToVisit[nextPosition++] = next;
              break;
            } 
            if (TypeBinding.equalsEquals(next, interfacesToVisit[b]))
              break; 
            b++;
          } 
        } 
      } 
    } 
    return false;
  }
  
  boolean implementsMethod(MethodBinding method) {
    char[] selector = method.selector;
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding type = this;
    while (type != null) {
      MethodBinding[] methods = type.methods();
      long range;
      if ((range = binarySearch(selector, methods)) >= 0L) {
        int start = (int)range, end = (int)(range >> 32L);
        for (int i = start; i <= end; i++) {
          if (methods[i].areParametersEqual(method))
            return true; 
        } 
      } 
      type = type.superclass();
    } 
    return false;
  }
  
  public final boolean isAbstract() {
    return ((this.modifiers & 0x400) != 0);
  }
  
  public boolean isAnnotationType() {
    return ((this.modifiers & 0x2000) != 0);
  }
  
  public final boolean isBinaryBinding() {
    return ((this.tagBits & 0x40L) != 0L);
  }
  
  public boolean isClass() {
    return ((this.modifiers & 0x6200) == 0);
  }
  
  public boolean isProperType(boolean admitCapture18) {
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding outer = enclosingType();
    if (outer != null && !outer.isProperType(admitCapture18))
      return false; 
    return super.isProperType(admitCapture18);
  }
  
  public boolean isCompatibleWith(TypeBinding otherType, Scope captureScope) {
    if (equalsEquals(otherType, this))
      return true; 
    if (otherType.id == 1)
      return true; 
    if (this.compatibleCache == null) {
      this.compatibleCache = new SimpleLookupTable(3);
      Object result = null;
    } else {
      Object result = this.compatibleCache.get(otherType);
      if (result != null)
        return (result == Boolean.TRUE); 
    } 
    this.compatibleCache.put(otherType, Boolean.FALSE);
    if (isCompatibleWith0(otherType, captureScope)) {
      this.compatibleCache.put(otherType, Boolean.TRUE);
      return true;
    } 
    if (captureScope == null && 
      this instanceof TypeVariableBinding && 
      ((TypeVariableBinding)this).firstBound instanceof org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding)
      this.compatibleCache.put(otherType, null); 
    return false;
  }
  
  private boolean isCompatibleWith0(TypeBinding otherType, Scope captureScope) {
    CaptureBinding otherCapture;
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding otherReferenceType;
    TypeBinding otherLowerBound;
    if (TypeBinding.equalsEquals(otherType, this))
      return true; 
    if (otherType.id == 1)
      return true; 
    if (isEquivalentTo(otherType))
      return true; 
    switch (otherType.kind()) {
      case 516:
      case 8196:
        return false;
      case 4100:
        otherCapture = (CaptureBinding)otherType;
        if (otherType.isCapture() && (otherLowerBound = otherCapture.lowerBound) != null) {
          if (otherLowerBound.isArrayType())
            return false; 
          return isCompatibleWith(otherLowerBound);
        } 
        if (otherType instanceof org.eclipse.jdt.internal.compiler.lookup.InferenceVariable)
          if (captureScope != null) {
            MethodScope methodScope = captureScope.methodScope();
            if (methodScope != null) {
              ReferenceContext referenceContext = methodScope.referenceContext;
              if (referenceContext instanceof LambdaExpression && 
                ((LambdaExpression)referenceContext).inferenceContext != null)
                return true; 
            } 
          }  
      case 4:
      case 260:
      case 1028:
      case 2052:
        switch (kind()) {
          case 260:
          case 1028:
          case 2052:
            if (TypeBinding.equalsEquals(erasure(), otherType.erasure()))
              return false; 
            break;
        } 
        otherReferenceType = (org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)otherType;
        if (otherReferenceType.isInterface()) {
          if (implementsInterface(otherReferenceType, true))
            return true; 
          if (this instanceof TypeVariableBinding && captureScope != null) {
            TypeVariableBinding typeVariable = (TypeVariableBinding)this;
            if (typeVariable.firstBound instanceof org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding) {
              TypeBinding bound = typeVariable.firstBound.capture(captureScope, -1, -1);
              return bound.isCompatibleWith(otherReferenceType);
            } 
          } 
        } 
        if (isInterface())
          return false; 
        return otherReferenceType.isSuperclassOf(this);
    } 
    return false;
  }
  
  public boolean isSubtypeOf(TypeBinding other) {
    if (isSubTypeOfRTL(other))
      return true; 
    TypeBinding candidate = findSuperTypeOriginatingFrom(other);
    if (candidate == null)
      return false; 
    if (TypeBinding.equalsEquals(candidate, other))
      return true; 
    if (other.isRawType() && TypeBinding.equalsEquals(candidate.erasure(), other.erasure()))
      return true; 
    TypeBinding[] sis = other.typeArguments();
    TypeBinding[] tis = candidate.typeArguments();
    if (tis == null || sis == null)
      return false; 
    if (sis.length != tis.length)
      return false; 
    for (int i = 0; i < sis.length; i++) {
      if (!tis[i].isTypeArgumentContainedBy(sis[i]))
        return false; 
    } 
    return true;
  }
  
  protected boolean isSubTypeOfRTL(TypeBinding other) {
    if (TypeBinding.equalsEquals(this, other))
      return true; 
    if (other instanceof CaptureBinding) {
      TypeBinding lower = ((CaptureBinding)other).lowerBound;
      return (lower != null && isSubtypeOf(lower));
    } 
    if (other instanceof org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding) {
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] arrayOfReferenceBinding = ((org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)other).getIntersectingTypes();
      if (arrayOfReferenceBinding != null) {
        for (int i = 0; i < arrayOfReferenceBinding.length; i++) {
          if (!isSubtypeOf(arrayOfReferenceBinding[i]))
            return false; 
        } 
        return true;
      } 
    } 
    return false;
  }
  
  public final boolean isDefault() {
    return ((this.modifiers & 0x7) == 0);
  }
  
  public final boolean isDeprecated() {
    return ((this.modifiers & 0x100000) != 0);
  }
  
  public boolean isEnum() {
    return ((this.modifiers & 0x4000) != 0);
  }
  
  public final boolean isFinal() {
    return ((this.modifiers & 0x10) != 0);
  }
  
  public boolean isHierarchyBeingConnected() {
    return ((this.tagBits & 0x200L) == 0L && (this.tagBits & 0x100L) != 0L);
  }
  
  public boolean isHierarchyBeingActivelyConnected() {
    return ((this.tagBits & 0x200L) == 0L && (this.tagBits & 0x100L) != 0L && (this.tagBits & 0x80000L) == 0L);
  }
  
  public boolean isHierarchyConnected() {
    return true;
  }
  
  public boolean isInterface() {
    return ((this.modifiers & 0x200) != 0);
  }
  
  public boolean isFunctionalInterface(Scope scope) {
    MethodBinding method;
    return (isInterface() && (method = getSingleAbstractMethod(scope, true)) != null && method.isValidBinding());
  }
  
  public final boolean isPrivate() {
    return ((this.modifiers & 0x2) != 0);
  }
  
  public final boolean isOrEnclosedByPrivateType() {
    if (isLocalType())
      return true; 
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding type = this;
    while (type != null) {
      if ((type.modifiers & 0x2) != 0)
        return true; 
      type = type.enclosingType();
    } 
    return false;
  }
  
  public final boolean isProtected() {
    return ((this.modifiers & 0x4) != 0);
  }
  
  public final boolean isPublic() {
    return ((this.modifiers & 0x1) != 0);
  }
  
  public final boolean isStatic() {
    return !((this.modifiers & 0x208) == 0 && (this.tagBits & 0x4L) != 0L);
  }
  
  public final boolean isStrictfp() {
    return ((this.modifiers & 0x800) != 0);
  }
  
  public boolean isSuperclassOf(org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding otherType) {
    while ((otherType = otherType.superclass()) != null) {
      if (otherType.isEquivalentTo(this))
        return true; 
    } 
    return false;
  }
  
  public boolean isThrowable() {
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding current = this;
    while (true) {
      switch (current.id) {
        case 19:
        case 21:
        case 24:
        case 25:
          return true;
      } 
      if ((current = current.superclass()) == null)
        return false; 
    } 
  }
  
  public boolean isUncheckedException(boolean includeSupertype) {
    switch (this.id) {
      case 19:
      case 24:
        return true;
      case 21:
      case 25:
        return includeSupertype;
    } 
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding current = this;
    while ((current = current.superclass()) != null) {
      switch (current.id) {
        case 19:
        case 24:
          return true;
        case 21:
        case 25:
          return false;
      } 
    } 
    return false;
  }
  
  public final boolean isUsed() {
    return ((this.modifiers & 0x8000000) != 0);
  }
  
  public final boolean isViewedAsDeprecated() {
    return !((this.modifiers & 0x300000) == 0 && 
      !getPackage().isViewedAsDeprecated());
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] memberTypes() {
    return Binding.NO_MEMBER_TYPES;
  }
  
  public MethodBinding[] methods() {
    return Binding.NO_METHODS;
  }
  
  public final org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding outermostEnclosingType() {
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding last, current = this;
    do {
      last = current;
    } while ((current = current.enclosingType()) != null);
    return last;
  }
  
  public char[] qualifiedSourceName() {
    if (isMemberType())
      return CharOperation.concat(enclosingType().qualifiedSourceName(), sourceName(), '.'); 
    return sourceName();
  }
  
  public char[] readableName() {
    char[] readableName;
    if (isMemberType()) {
      readableName = CharOperation.concat(enclosingType().readableName(), this.sourceName, '.');
    } else {
      readableName = CharOperation.concatWith(this.compoundName, '.');
    } 
    TypeVariableBinding[] typeVars;
    if ((typeVars = typeVariables()) != Binding.NO_TYPE_VARIABLES) {
      StringBuffer nameBuffer = new StringBuffer(10);
      nameBuffer.append(readableName).append('<');
      for (int i = 0, length = typeVars.length; i < length; i++) {
        if (i > 0)
          nameBuffer.append(','); 
        nameBuffer.append(typeVars[i].readableName());
      } 
      nameBuffer.append('>');
      int nameLength = nameBuffer.length();
      readableName = new char[nameLength];
      nameBuffer.getChars(0, nameLength, readableName, 0);
    } 
    return readableName;
  }
  
  protected void appendNullAnnotation(StringBuffer nameBuffer, CompilerOptions options) {
    if (options.isAnnotationBasedNullAnalysisEnabled)
      if (options.usesNullTypeAnnotations()) {
        byte b;
        int i;
        AnnotationBinding[] arrayOfAnnotationBinding;
        for (i = (arrayOfAnnotationBinding = this.typeAnnotations).length, b = 0; b < i; ) {
          AnnotationBinding annotation = arrayOfAnnotationBinding[b];
          org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding annotationType = annotation.getAnnotationType();
          if (annotationType.hasNullBit(96))
            nameBuffer.append('@').append(annotationType.shortReadableName()).append(' '); 
          b++;
        } 
      } else {
        if ((this.tagBits & 0x100000000000000L) != 0L) {
          char[][] nonNullAnnotationName = options.nonNullAnnotationName;
          nameBuffer.append('@').append(nonNullAnnotationName[nonNullAnnotationName.length - 1]).append(' ');
        } 
        if ((this.tagBits & 0x80000000000000L) != 0L) {
          char[][] nullableAnnotationName = options.nullableAnnotationName;
          nameBuffer.append('@').append(nullableAnnotationName[nullableAnnotationName.length - 1]).append(' ');
        } 
      }  
  }
  
  public AnnotationHolder retrieveAnnotationHolder(Binding binding, boolean forceInitialization) {
    SimpleLookupTable store = storedAnnotations(forceInitialization);
    return (store == null) ? null : (AnnotationHolder)store.get(binding);
  }
  
  AnnotationBinding[] retrieveAnnotations(Binding binding) {
    AnnotationHolder holder = retrieveAnnotationHolder(binding, true);
    return (holder == null) ? Binding.NO_ANNOTATIONS : holder.getAnnotations();
  }
  
  public void setAnnotations(AnnotationBinding[] annotations) {
    storeAnnotations((Binding)this, annotations);
  }
  
  public void setContainerAnnotationType(org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding value) {}
  
  public void tagAsHavingDefectiveContainerType() {}
  
  public char[] nullAnnotatedReadableName(CompilerOptions options, boolean shortNames) {
    if (shortNames)
      return nullAnnotatedShortReadableName(options); 
    return nullAnnotatedReadableName(options);
  }
  
  char[] nullAnnotatedReadableName(CompilerOptions options) {
    StringBuffer nameBuffer = new StringBuffer(10);
    if (isMemberType()) {
      nameBuffer.append(enclosingType().nullAnnotatedReadableName(options, false));
      nameBuffer.append('.');
      appendNullAnnotation(nameBuffer, options);
      nameBuffer.append(this.sourceName);
    } else if (this.compoundName != null) {
      int l = this.compoundName.length;
      int i;
      for (i = 0; i < l - 1; i++) {
        nameBuffer.append(this.compoundName[i]);
        nameBuffer.append('.');
      } 
      appendNullAnnotation(nameBuffer, options);
      nameBuffer.append(this.compoundName[i]);
    } else {
      appendNullAnnotation(nameBuffer, options);
      if (this.sourceName != null) {
        nameBuffer.append(this.sourceName);
      } else {
        nameBuffer.append(readableName());
      } 
    } 
    TypeBinding[] arguments = typeArguments();
    if (arguments != null && arguments.length > 0) {
      nameBuffer.append('<');
      for (int i = 0, length = arguments.length; i < length; i++) {
        if (i > 0)
          nameBuffer.append(','); 
        nameBuffer.append(arguments[i].nullAnnotatedReadableName(options, false));
      } 
      nameBuffer.append('>');
    } 
    int nameLength = nameBuffer.length();
    char[] readableName = new char[nameLength];
    nameBuffer.getChars(0, nameLength, readableName, 0);
    return readableName;
  }
  
  char[] nullAnnotatedShortReadableName(CompilerOptions options) {
    StringBuffer nameBuffer = new StringBuffer(10);
    if (isMemberType()) {
      nameBuffer.append(enclosingType().nullAnnotatedReadableName(options, true));
      nameBuffer.append('.');
      appendNullAnnotation(nameBuffer, options);
      nameBuffer.append(this.sourceName);
    } else {
      appendNullAnnotation(nameBuffer, options);
      if (this.sourceName != null) {
        nameBuffer.append(this.sourceName);
      } else {
        nameBuffer.append(shortReadableName());
      } 
    } 
    TypeBinding[] arguments = typeArguments();
    if (arguments != null && arguments.length > 0) {
      nameBuffer.append('<');
      for (int i = 0, length = arguments.length; i < length; i++) {
        if (i > 0)
          nameBuffer.append(','); 
        nameBuffer.append(arguments[i].nullAnnotatedReadableName(options, true));
      } 
      nameBuffer.append('>');
    } 
    int nameLength = nameBuffer.length();
    char[] shortReadableName = new char[nameLength];
    nameBuffer.getChars(0, nameLength, shortReadableName, 0);
    return shortReadableName;
  }
  
  public char[] shortReadableName() {
    char[] shortReadableName;
    if (isMemberType()) {
      shortReadableName = CharOperation.concat(enclosingType().shortReadableName(), this.sourceName, '.');
    } else {
      shortReadableName = this.sourceName;
    } 
    TypeVariableBinding[] typeVars;
    if ((typeVars = typeVariables()) != Binding.NO_TYPE_VARIABLES) {
      StringBuffer nameBuffer = new StringBuffer(10);
      nameBuffer.append(shortReadableName).append('<');
      for (int i = 0, length = typeVars.length; i < length; i++) {
        if (i > 0)
          nameBuffer.append(','); 
        nameBuffer.append(typeVars[i].shortReadableName());
      } 
      nameBuffer.append('>');
      int nameLength = nameBuffer.length();
      shortReadableName = new char[nameLength];
      nameBuffer.getChars(0, nameLength, shortReadableName, 0);
    } 
    return shortReadableName;
  }
  
  public char[] signature() {
    if (this.signature != null)
      return this.signature; 
    return this.signature = CharOperation.concat('L', constantPoolName(), ';');
  }
  
  public char[] sourceName() {
    return this.sourceName;
  }
  
  void storeAnnotationHolder(Binding binding, AnnotationHolder holder) {
    if (holder == null) {
      SimpleLookupTable store = storedAnnotations(false);
      if (store != null)
        store.removeKey(binding); 
    } else {
      SimpleLookupTable store = storedAnnotations(true);
      if (store != null)
        store.put(binding, holder); 
    } 
  }
  
  void storeAnnotations(Binding binding, AnnotationBinding[] annotations) {
    AnnotationHolder holder = null;
    if (annotations == null || annotations.length == 0) {
      SimpleLookupTable store = storedAnnotations(false);
      if (store != null)
        holder = (AnnotationHolder)store.get(binding); 
      if (holder == null)
        return; 
    } else {
      SimpleLookupTable store = storedAnnotations(true);
      if (store == null)
        return; 
      holder = (AnnotationHolder)store.get(binding);
      if (holder == null)
        holder = new AnnotationHolder(); 
    } 
    storeAnnotationHolder(binding, holder.setAnnotations(annotations));
  }
  
  SimpleLookupTable storedAnnotations(boolean forceInitialize) {
    return null;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding superclass() {
    return null;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] superInterfaces() {
    return Binding.NO_SUPERINTERFACES;
  }
  
  public org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] syntheticEnclosingInstanceTypes() {
    if (isStatic())
      return null; 
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding enclosingType = enclosingType();
    if (enclosingType == null)
      return null; 
    return new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] { enclosingType };
  }
  
  MethodBinding[] unResolvedMethods() {
    return methods();
  }
  
  public FieldBinding[] unResolvedFields() {
    return Binding.NO_FIELDS;
  }
  
  protected int applyCloseableClassWhitelists() {
    switch (this.compoundName.length) {
      case 3:
        if (CharOperation.equals(TypeConstants.JAVA, this.compoundName[0]) && 
          CharOperation.equals(TypeConstants.IO, this.compoundName[1])) {
          char[] simpleName = this.compoundName[2];
          int j = TypeConstants.JAVA_IO_WRAPPER_CLOSEABLES.length;
          int k;
          for (k = 0; k < j; k++) {
            if (CharOperation.equals(simpleName, TypeConstants.JAVA_IO_WRAPPER_CLOSEABLES[k]))
              return 4; 
          } 
          j = TypeConstants.JAVA_IO_RESOURCE_FREE_CLOSEABLES.length;
          for (k = 0; k < j; k++) {
            if (CharOperation.equals(simpleName, TypeConstants.JAVA_IO_RESOURCE_FREE_CLOSEABLES[k]))
              return 8; 
          } 
        } 
        break;
      case 4:
        if (CharOperation.equals(TypeConstants.JAVA, this.compoundName[0]) && 
          CharOperation.equals(TypeConstants.UTIL, this.compoundName[1]) && 
          CharOperation.equals(TypeConstants.ZIP, this.compoundName[2])) {
          char[] simpleName = this.compoundName[3];
          int j = TypeConstants.JAVA_UTIL_ZIP_WRAPPER_CLOSEABLES.length;
          for (int k = 0; k < j; k++) {
            if (CharOperation.equals(simpleName, TypeConstants.JAVA_UTIL_ZIP_WRAPPER_CLOSEABLES[k]))
              return 4; 
          } 
        } 
        break;
    } 
    int l = TypeConstants.OTHER_WRAPPER_CLOSEABLES.length;
    for (int i = 0; i < l; i++) {
      if (CharOperation.equals(this.compoundName, TypeConstants.OTHER_WRAPPER_CLOSEABLES[i]))
        return 4; 
    } 
    return 0;
  }
  
  protected int applyCloseableInterfaceWhitelists() {
    switch (this.compoundName.length) {
      case 4:
        if (CharOperation.equals(this.compoundName, TypeConstants.RESOURCE_FREE_CLOSEABLE_STREAM))
          return 8; 
        break;
    } 
    return 0;
  }
  
  protected MethodBinding[] getInterfaceAbstractContracts(Scope scope, boolean replaceWildcards) throws InvalidInputException {
    if (!isInterface() || !isValidBinding())
      throw new InvalidInputException("Not a functional interface"); 
    MethodBinding[] methods = methods();
    MethodBinding[] contracts = new MethodBinding[0];
    int contractsCount = 0;
    int contractsLength = 0;
    org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] superInterfaces = superInterfaces();
    int i, length;
    for (i = 0, length = superInterfaces.length; i < length; i++) {
      MethodBinding[] superInterfaceContracts = superInterfaces[i].getInterfaceAbstractContracts(scope, replaceWildcards);
      int superInterfaceContractsLength = (superInterfaceContracts == null) ? 0 : superInterfaceContracts.length;
      if (superInterfaceContractsLength != 0) {
        if (contractsLength < contractsCount + superInterfaceContractsLength)
          System.arraycopy(contracts, 0, contracts = new MethodBinding[contractsLength = contractsCount + superInterfaceContractsLength], 0, contractsCount); 
        System.arraycopy(superInterfaceContracts, 0, contracts, contractsCount, superInterfaceContractsLength);
        contractsCount += superInterfaceContractsLength;
      } 
    } 
    for (i = 0, length = (methods == null) ? 0 : methods.length; i < length; i++) {
      MethodBinding method = methods[i];
      if (method != null && !method.isStatic() && !method.redeclaresPublicObjectMethod(scope)) {
        if (!method.isValidBinding())
          throw new InvalidInputException("Not a functional interface"); 
        for (int j = 0; j < contractsCount; ) {
          if (contracts[j] != null && MethodVerifier.doesMethodOverride(method, contracts[j], scope.environment())) {
            contractsCount--;
            if (j < contractsCount) {
              System.arraycopy(contracts, j + 1, contracts, j, contractsCount - j);
              continue;
            } 
          } 
          j++;
        } 
        if (!method.isDefaultMethod()) {
          if (contractsCount == contractsLength) {
            contractsLength += 16;
            System.arraycopy(contracts, 0, contracts = new MethodBinding[contractsLength], 0, contractsCount);
          } 
          contracts[contractsCount++] = method;
        } 
      } 
    } 
    if (contractsCount < contractsLength)
      System.arraycopy(contracts, 0, contracts = new MethodBinding[contractsCount], 0, contractsCount); 
    return contracts;
  }
  
  public MethodBinding getSingleAbstractMethod(Scope scope, boolean replaceWildcards) {
    int index = replaceWildcards ? 0 : 1;
    if (this.singleAbstractMethod != null) {
      if (this.singleAbstractMethod[index] != null)
        return this.singleAbstractMethod[index]; 
    } else {
      this.singleAbstractMethod = new MethodBinding[2];
    } 
    if (this.compoundName != null)
      scope.compilationUnitScope().recordQualifiedReference(this.compoundName); 
    MethodBinding[] methods = null;
    try {
      methods = getInterfaceAbstractContracts(scope, replaceWildcards);
      if (methods == null || methods.length == 0) {
        this.singleAbstractMethod[index] = (MethodBinding)samProblemBinding;
        return (MethodBinding)samProblemBinding;
      } 
      int contractParameterLength = 0;
      char[] contractSelector = null;
      for (int j = 0, k = methods.length; j < k; j++) {
        MethodBinding method = methods[j];
        if (method != null)
          if (contractSelector == null) {
            contractSelector = method.selector;
            contractParameterLength = (method.parameters == null) ? 0 : method.parameters.length;
          } else {
            int methodParameterLength = (method.parameters == null) ? 0 : method.parameters.length;
            if (methodParameterLength != contractParameterLength || !CharOperation.equals(method.selector, contractSelector)) {
              this.singleAbstractMethod[index] = (MethodBinding)samProblemBinding;
              return (MethodBinding)samProblemBinding;
            } 
          }  
      } 
    } catch (InvalidInputException invalidInputException) {
      this.singleAbstractMethod[index] = (MethodBinding)samProblemBinding;
      return (MethodBinding)samProblemBinding;
    } 
    if (methods.length == 1) {
      this.singleAbstractMethod[index] = methods[0];
      return methods[0];
    } 
    LookupEnvironment environment = scope.environment();
    boolean genericMethodSeen = false;
    int length = methods.length;
    boolean analyseNullAnnotations = environment.globalOptions.isAnnotationBasedNullAnalysisEnabled;
    for (int i = length - 1; i >= 0; ) {
      MethodBinding method = methods[i], otherMethod = null;
      if (method.typeVariables != Binding.NO_TYPE_VARIABLES)
        genericMethodSeen = true; 
      TypeBinding returnType = method.returnType;
      TypeBinding[] parameters = method.parameters;
      for (int j = 0; j < length; j++) {
        if (i != j) {
          otherMethod = methods[j];
          if (otherMethod.typeVariables != Binding.NO_TYPE_VARIABLES)
            genericMethodSeen = true; 
          if (genericMethodSeen) {
            otherMethod = MethodVerifier.computeSubstituteMethod(otherMethod, method, environment);
            if (otherMethod == null)
              // Byte code: goto -> 985 
          } 
          if (MethodVerifier.isSubstituteParameterSubsignature(method, otherMethod, environment)) {
            if (!MethodVerifier.areReturnTypesCompatible(method, otherMethod, environment)) {
              i--;
              continue;
            } 
            if (analyseNullAnnotations) {
              returnType = NullAnnotationMatching.strongerType(returnType, otherMethod.returnType, environment);
              parameters = NullAnnotationMatching.weakerTypes(parameters, otherMethod.parameters, environment);
            } 
          } else {
            // Byte code: goto -> 985
          } 
        } 
      } 
      org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] exceptions = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[0];
      int exceptionsCount = 0, exceptionsLength = 0;
      MethodBinding theAbstractMethod = method;
      boolean shouldEraseThrows = (theAbstractMethod.typeVariables == Binding.NO_TYPE_VARIABLES && genericMethodSeen);
      boolean shouldAdaptThrows = (theAbstractMethod.typeVariables != Binding.NO_TYPE_VARIABLES);
      int typeVariableLength = theAbstractMethod.typeVariables.length;
      label154: for (i = 0; i < length; i++) {
        method = methods[i];
        org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] methodThrownExceptions = method.thrownExceptions;
        int methodExceptionsLength = (methodThrownExceptions == null) ? 0 : methodThrownExceptions.length;
        if (methodExceptionsLength == 0)
          break; 
        if (shouldAdaptThrows && method != theAbstractMethod) {
          System.arraycopy(methodThrownExceptions, 0, methodThrownExceptions = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[methodExceptionsLength], 0, methodExceptionsLength);
          for (int tv = 0; tv < typeVariableLength; tv++) {
            if (methodThrownExceptions[tv] instanceof TypeVariableBinding)
              methodThrownExceptions[tv] = (org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)theAbstractMethod.typeVariables[tv]; 
          } 
        } 
        for (int k = 0; k < methodExceptionsLength; k++) {
          org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding methodException = methodThrownExceptions[k];
          if (shouldEraseThrows)
            methodException = (org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)methodException.erasure(); 
          for (int m = 0; m < length; ) {
            if (i == m) {
              m++;
              continue;
            } 
            otherMethod = methods[m];
            org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[] otherMethodThrownExceptions = otherMethod.thrownExceptions;
            int otherMethodExceptionsLength = (otherMethodThrownExceptions == null) ? 0 : otherMethodThrownExceptions.length;
            if (otherMethodExceptionsLength == 0)
              break label154; 
            if (shouldAdaptThrows && otherMethod != theAbstractMethod) {
              System.arraycopy(otherMethodThrownExceptions, 
                  0, 
                  otherMethodThrownExceptions = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[otherMethodExceptionsLength], 
                  0, 
                  otherMethodExceptionsLength);
              for (int tv = 0; tv < typeVariableLength; tv++) {
                if (otherMethodThrownExceptions[tv] instanceof TypeVariableBinding)
                  otherMethodThrownExceptions[tv] = (org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)theAbstractMethod.typeVariables[tv]; 
              } 
            } 
            for (int l = 0; l < otherMethodExceptionsLength; l++) {
              org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding otherException = otherMethodThrownExceptions[l];
              if (shouldEraseThrows)
                otherException = (org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding)otherException.erasure(); 
              if (methodException.isCompatibleWith(otherException))
                // Byte code: goto -> 835 
            } 
            // Byte code: goto -> 882
          } 
          if (exceptionsCount == exceptionsLength) {
            exceptionsLength += 16;
            System.arraycopy(exceptions, 0, exceptions = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[exceptionsLength], 0, exceptionsCount);
          } 
          exceptions[exceptionsCount++] = methodException;
        } 
      } 
      if (exceptionsCount != exceptionsLength)
        System.arraycopy(exceptions, 0, exceptions = new org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding[exceptionsCount], 0, exceptionsCount); 
      this.singleAbstractMethod[index] = new MethodBinding(theAbstractMethod.modifiers | 0x1000, 
          theAbstractMethod.selector, 
          returnType, 
          parameters, 
          exceptions, 
          theAbstractMethod.declaringClass);
      (this.singleAbstractMethod[index]).typeVariables = theAbstractMethod.typeVariables;
      return this.singleAbstractMethod[index];
    } 
    this.singleAbstractMethod[index] = (MethodBinding)samProblemBinding;
    return (MethodBinding)samProblemBinding;
  }
  
  public static boolean isConsistentIntersection(TypeBinding[] intersectingTypes) {
    TypeBinding[] ci = new TypeBinding[intersectingTypes.length];
    for (int i = 0; i < ci.length; i++) {
      TypeBinding current = intersectingTypes[i];
      ci[i] = (current.isClass() || current.isArrayType()) ? 
        current : current.superclass();
    } 
    TypeBinding mostSpecific = ci[0];
    for (int j = 1; j < ci.length; j++) {
      TypeBinding current = ci[j];
      if (!current.isTypeVariable() && !current.isWildcard() && current.isProperType(true))
        if (!mostSpecific.isSubtypeOf(current))
          if (current.isSubtypeOf(mostSpecific)) {
            mostSpecific = current;
          } else {
            return false;
          }   
    } 
    return true;
  }
}
